@@ -7,16 +7,17 @@ use warnings FATAL => 'all';
 use Exporter 'import';
 
 our @EXPORT = qw( );
-our @EXPORT_OK = qw( readRDS readRData );
+our @EXPORT_OK = qw( readRDS readRData evalRserve );
 
 our %EXPORT_TAGS = ( all => [ @EXPORT_OK ], );
 
 use Statistics::R::IO::REXPFactory;
 use IO::Uncompress::Gunzip ();
 use IO::Uncompress::Bunzip2 ();
+use IO::Socket::INET ();
 use Carp;
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 
 sub readRDS {
@@ -73,6 +74,66 @@ sub readRData {
     Statistics::R::IO::REXPFactory::tagged_pairlist_to_rexp_hash $value;
 }
 
+
+sub evalRserve {
+    my ($rexp, $server) = (shift, shift);
+
+    my $fh;
+    if (UNIVERSAL::isa($server, 'IO::Handle')) {
+        $fh = $server;
+    }
+    else {
+        $server //= 'localhost';
+        my $port = shift || 6311;
+        $fh = IO::Socket::INET->new(PeerAddr => $server,
+                                    PeerPort => $port) or
+                                        croak $!;
+        $fh->sysread(my $response, 32);
+        croak "Unrecognized server ID" unless
+            substr($response, 0, 12) eq 'Rsrv0103QAP1';
+    }
+
+    ## simulate the request parameter as constructed by:
+    ## > serialize(quote(parse(text="{$rexp}")[[1]]), NULL)
+    my $parameter =
+        "\x58\x0a\0\0\0\2\0\3\0\3\0\2\3\0\0\0\0\6\0\0\0\1\0\4\0" .
+        "\x09\0\0\0\2\x5b\x5b\0\0\0\2\0\0\0\6\0\0\0\1\0\4\0\x09\0\0" .
+        "\0\5\x70\x61\x72\x73\x65\0\0\4\2\0\0\0\1\0\4\0\x09\0\0\0\4\x74\x65" .
+        "\x78\x74\0\0\0\x10\0\0\0\1\0\4\0\x09" .
+        pack('N', length($rexp)+2) .
+        "\x7b" . $rexp . "\x7d" .
+        "\0\0\0\xfe\0\0\0\2\0\0\0\x0e\0\0\0\1\x3f\xf0\0\0\0\0\0\0" .
+        "\0\0\0\xfe";
+    ## request is:
+    ## - command (0xf5, CMD_serEval,
+    ##       means raw serialized data without data header)
+    ## - length of the message (low 32 bits)
+    ## - offset of the data part
+    ## - high 32 bits of the length of the message (0 if < 4GB)
+    $fh->syswrite(pack('V4', 245, length($parameter), 0, 0) .
+                  $parameter);
+    
+    $fh->sysread(my $response, 16);
+    ## Of the next four long-ints:
+    ## - the first one is status and should be 65537 (bytes \1, \0, \1, \0)
+    ## - the second one is length
+    ## - the third and fourth are ??
+    my ($status, $length) = unpack VV => substr($response, 0, 8);
+    unless ($status == 65537) {
+        croak 'Server returned an error: ' . $status;
+    }
+    
+    $fh->sysread(my $data, $length);
+
+    ## Close the handle only if it was created in this function
+    close $fh unless $server eq $fh;
+    
+    my ($value, $state) = @{Statistics::R::IO::REXPFactory::unserialize($data)};
+    croak 'Could not parse Rserve value' unless $state;
+    croak 'Unread data remaining in the Rserve response' unless $state->eof;
+    $value
+}
+
 1; # End of Statistics::R::IO
 
 __END__
@@ -85,7 +146,7 @@ Statistics::R::IO - Perl interface to serialized R data
 
 =head1 VERSION
 
-This documentation refers to version 0.03 of the module.
+This documentation refers to version 0.04 of the module.
 
 
 =head1 SYNOPSIS
@@ -100,6 +161,8 @@ This documentation refers to version 0.03 of the module.
         print $var_name, $value;
     }
 
+    my $pi = Statistics::R::IO::evalRserve('pi');
+    print $pi->to_pl;
 
 =head1 DESCRIPTION
 
@@ -131,6 +194,11 @@ on startup.)
 
 =back
 
+As of version 0.04, the module can also evaluate R code on a remote
+host that runs the L<Rserve|http://www.rforge.net/Rserve/> binary R
+server. This allows Perl programs to access all facilities of R
+without the need to have a local install of R or link to an R library.
+
 See L</SUBROUTINES> for invocation and usage information on individual
 subroutines, and the L<R Internals
 manual|http://cran.r-project.org/doc/manuals/R-ints.html> for the
@@ -139,8 +207,8 @@ specification of the file formats.
 
 =head1 EXPORT
 
-Nothing by default. Optionally, subroutines C<readRDS> and
-C<readRData>, or C<:all> for both.
+Nothing by default. Optionally, subroutines C<readRDS>, C<readRData>,
+and C<evalRserve>, or C<:all> for all three.
 
 
 =head1 SUBROUTINES
@@ -158,6 +226,26 @@ Reads a file in RData format whose filename is given by EXPR and
 returns a hash whose keys are the names of objects stored in the file
 with corresponding values as L<Statistics::R::REXP> instances.
 
+=item evalRserve REXPR [ HOSTNAME [, PORT] | HANDLE]
+
+Evaluates an R expression, given as text string in REXPR, on an
+L<Rserve|http://www.rforge.net/Rserve/> server and returns its result
+as a L<Statistics::R::REXP> object.
+
+The server location can be specified either by its host name and
+(optionally) port or by a connected instance of L<IO::Handle>. The
+caller passing the HANDLE is responsible for reading (and checking)
+the server ID that is returned in the first 32-byte response when the
+connection was established. This allows opening the connection once
+and reusing it in multiple calls to 'evalRserve'.
+
+If only REXPR is given, the function assumes that the server runs on
+the localhost. If PORT is not specified, it defaults to the standard
+Rserve port, 6311.
+
+The function will close the connection to the Rserve host if it has
+opened it itself, but not if the connection was passed as a HANDLE.
+
 =back
 
 
